from click.testing import CliRunner
from choixe.cli.compile import compile
from pathlib import Path


def test_compile(sample_configurations_data, tmpdir):
    runner = CliRunner()

    for cfg_data in sample_configurations_data:
        filename = cfg_data['filename']
        has_placeholders = cfg_data['has_placeholders']
        to_replace = cfg_data['to_replace'] if 'to_replace' in cfg_data else []

        outfile = Path(tmpdir) / f'_copy_{Path(filename).name}'

        if has_placeholders:
            result = runner.invoke(compile, ['-c', filename, '-o', outfile])
            assert result.exit_code == 1

            result = runner.invoke(compile, ['-c', filename, '-o', outfile, '--nocheck'])
            assert result.exit_code == 0
        else:
            result = runner.invoke(compile, ['-c', filename, '-o', outfile])
            assert result.exit_code == 0

            result = runner.invoke(compile, ['-c', filename, '-o', outfile, '--nocheck'])
            assert result.exit_code == 0

        if len(to_replace) > 0 and has_placeholders:
            args = ['-c', filename, '-o', outfile]
            for k, v in to_replace.items():
                args += ['--option', k, v]
            result = runner.invoke(compile, args)
            assert result.exit_code == 0

        # result = runner.invoke(check, ['-c', cfg_file, '--noclose'])
        # assert result.exit_code == 0


def test_compile_wrong(sample_configurations_data, tmpdir):
    runner = CliRunner()

    for cfg_data in sample_configurations_data:
        filename = Path(str(cfg_data['filename']) + "@IMPOSSIBLE_ST3ING!")
        has_placeholders = cfg_data['has_placeholders']

        outfile = Path(tmpdir) / f'_copy_{Path(filename).name}'

        result = runner.invoke(compile, ['-c', filename, '-o', outfile])
        assert result.exit_code == 1



import pytest
import os
from pathlib import Path


def pytest_addoption(parser):
    parser.addoption("--rabbitmq", action="store_true",
                     help="run the tests only in case of rabbitmq server active")


def pytest_runtest_setup(item):
    if 'rabbitmq' in item.keywords and not item.config.getoption("--rabbitmq"):
        pytest.skip("need --rabbitmq for full testing with rabbitmq server")


@pytest.fixture(scope='session')
def data_folder():
    dirname = os.path.dirname(__file__)
    return os.path.join(dirname, 'sample_data')


@pytest.fixture(scope='session')
def sample_configurations_data(data_folder):
    configurations_folder = Path(data_folder) / 'configurations'
    return [
        {
            'filename': configurations_folder / 'sample_yaml' / 'main.yaml',
            'has_placeholders': True
        },
        {
            'filename': configurations_folder / 'sample_yaml' / 'main_single.yaml',
            'has_placeholders': True
        },
        {
            'filename': configurations_folder / 'single_yaml' / 'main_noplaceholders.yaml',
            'has_placeholders': False
        },
        {
            'filename': configurations_folder / 'single_yaml' / 'main.yaml',
            'has_placeholders': True,
            'to_replace': {  # If to_replace is set, it has to contain all placeholders!
                'one': 1,
                'four': 33,
                'three.t31': True,
                'three.name': 'hello'
            }
        }
    ]



import os
import rich
from choixe.directives import DirectiveAT
from choixe.placeholders import Placeholder
from typing import Sequence, Union
import copy
from box.box_list import BoxList
import numpy as np
from box.box import Box
from deepdiff import DeepDiff
from schema import Schema, Or, Regex, SchemaMissingKeyError

import pydash
from pathlib import Path
import pytest
from choixe.configurations import XConfig


@pytest.fixture(scope="function")
def generic_temp_folder(tmpdir_factory):
    fn = tmpdir_factory.mktemp("_choixe_temp_folder")
    return fn


def simple_data():
    sample_dict = {
        'one': 1,
        'two': np.random.randint(-50, 50, (3, 3)).tolist(),
        'three': {
            'a': 2,
            'c': True
        }
    }
    schema = Schema(
        {
            'one': Or(int, str),
            'two': list,
            'three': {
                'a': int,
                'c': bool
            }
        }
    )

    to_be_raplaced_keys = sorted([
        'three'
    ], reverse=True)

    return {
        'data': sample_dict,
        'schema': schema,
        'placeholders': [],
        'environment_variables': [],
        'to_be_replaced': to_be_raplaced_keys
    }


def complex_data():
    np.random.seed(666)

    env_variables = [
        generate_placeholder('v_10', 'env'),
        generate_placeholder('v_11', 'env'),
    ]

    placeholders = [
        generate_placeholder('v_0', 'int', [6, 7, 8, 'default=10']),
        generate_placeholder('v_1', 'float'),
        generate_placeholder('v_2', 'str', ['alpha', 'beta', 'gamma']),
        generate_placeholder('v_3', 'bool'),
        generate_placeholder('v_4'),
        generate_placeholder('v_5', 'path'),
        generate_placeholder('v_6', 'date'),
        generate_placeholder('v_7', 'str', ['A', 'B', 'C', 'default=C']),
        generate_placeholder('v_8'),
        generate_placeholder('v_9'),
        generate_placeholder('v_with_dots_0'),
        generate_placeholder('v_with_dots_1'),
        generate_placeholder('v_with_dots_2'),
        generate_placeholder('v_with_dots_3')
    ]

    sample_dict = {

        'one': placeholders[0],
        'two': np.random.randint(-50, 50, (3, 3)).tolist(),
        'three': {
            '3.1': 'TrueValue',
            '2.1': [False, False],

            'name': {
                'name_1': placeholders[1],
                'name_2': 2,
                'name_3': {
                    'this_is_a_list': [3.3, 3.3],
                    'this_is_A_dict': {
                        'a': {
                            'f': True,
                            's': False,
                            'numpy_array': np.array([1., 2, 3]),
                            'numpy_data': np.array([1., 2, 3])[0],
                            'tuple': ('a', 'b', '2'),
                            'boxlist': BoxList([1, 2, 3, 4]),
                            'boxdict': Box({'a': 2.2})
                        }
                    }
                }
            }
        },
        'first': {
            'f1': placeholders[2],
            'f2': 2.22,
            'f3': [3, 3, 3, 3, 3, 3],
            'external': {
                'ext': np.random.uniform(-2, 2, (2, 3)).tolist(),
                'ext_name': placeholders[3],
            },
            'ops': {
                'o1': placeholders[4],
                'o2': placeholders[5],
                'o3': placeholders[6],
                'o4': placeholders[7],
                'o5': placeholders[8],
                'o6': placeholders[9]
            },
            'env': [
                env_variables[0],
                env_variables[1]
            ],
            'key': {
                'with': 120,
                'dots': placeholders[12],
            },
            'key.with.dots': placeholders[10],
            'key.with...many.....dots': placeholders[11],
            'nested.key': {
                'with.dots': placeholders[13]
            }
        }
    }

    schema = Schema(
        {
            'one': Or(int, str),
            'two': list,
            'three': {
                '3.1': str,
                '2.1': [bool],
                'name': dict
            },
            'first': {
                Regex(''): Or(str, int, list, float, dict)
            }
        }
    )

    to_be_raplaced_keys = sorted([
        'three.name',
        'three.name.name_3',
        'three.name.name_3.this_is_A_dict',
        'first',
        'first.external'
    ], reverse=True)

    return {
        'data': XConfig.decode(sample_dict),
        'schema': schema,
        'placeholders': placeholders,
        'environment_variables': env_variables,
        'to_be_replaced': to_be_raplaced_keys
    }


def generate_placeholder(name: str, tp: str = None, others: Sequence = None):
    args = [name]
    if others is not None:
        args.extend(others)
    if tp is not None:
        if len(tp) > 0:
            return DirectiveAT.generate_directive_string(tp, args)
    return DirectiveAT.generate_directive_string('str', args)


def data_to_test():
    return [
        simple_data(),
        complex_data()
    ]


def store_cfg(filename: Union[str, Path], d: dict):

    filename = str(filename)
    data = Box(d)
    if 'yml' in filename or 'yaml' in filename:
        Box(XConfig.decode(data.to_dict())).to_yaml(filename)
    if 'json' in filename:
        Box(XConfig.decode(data.to_dict())).to_json(filename)
    if 'toml' in filename:
        Box(XConfig.decode(data.to_dict())).to_toml(filename)


class TestXConfig(object):

    @pytest.mark.parametrize("cfg_extension", ['yaml', 'json', 'toml'])
    @pytest.mark.parametrize("data", data_to_test())
    def test_creation(self, generic_temp_folder, data, cfg_extension):  # TODO: toml 0.10.2 needed! toml has a bug otherwise

        generic_temp_folder = Path(generic_temp_folder)

        sample_dict = data['data']
        to_be_raplaced_keys = data['to_be_replaced']  # , _, _, to_be_raplaced_keys = data
        volatile_dict = copy.deepcopy(sample_dict)

        subtitutions_values = {}
        for k in to_be_raplaced_keys:
            random_name = k + f".{cfg_extension}"  # {XConfig.REFERENCE_QUALIFIER}"
            random_name = DirectiveAT.generate_directive_string('import', [random_name])
            subtitutions_values[random_name] = pydash.get(volatile_dict, k)

            pydash.set_(volatile_dict, k, random_name)

        output_cfg_filename = generic_temp_folder / f'out_config.{cfg_extension}'
        output_cfg_filename2 = generic_temp_folder / f'out_config2.{cfg_extension}'

        subtitutions_values[str(output_cfg_filename)] = volatile_dict

        saved_cfgs = []
        for directive_value, d in subtitutions_values.items():
            directive = DirectiveAT(value=directive_value)
            if directive.valid:
                output_filename = generic_temp_folder / directive.args[0]
            else:
                output_filename = generic_temp_folder / directive_value
            store_cfg(output_filename, d)
            saved_cfgs.append(output_filename)

        yconf = XConfig(output_cfg_filename)
        yconf.save_to(output_cfg_filename2)

        yconf_reloaded = XConfig(output_cfg_filename2)

        with pytest.raises(NotImplementedError):
            yconf.save_to(str(output_cfg_filename2) + "#IMPOSSIBLE.EXTENSION")

        assert not DeepDiff(yconf.to_dict(), sample_dict)
        assert not DeepDiff(yconf_reloaded.to_dict(), sample_dict)
        assert not DeepDiff(XConfig.from_dict(yconf_reloaded.to_dict()).to_dict(), yconf_reloaded.to_dict())
        assert len(XConfig.from_dict(yconf_reloaded.to_dict())) > len(sample_dict)  # YConf contains 2 more private keys!

        # remove last cfg file
        saved_cfgs[0].unlink()

        with pytest.raises(OSError):
            yconf = XConfig(output_cfg_filename)

    @pytest.mark.parametrize("cfg_extension", ['yaml', 'json', 'toml'])
    @pytest.mark.parametrize("data", data_to_test())
    def test_creation_root_replace(self, generic_temp_folder, data, cfg_extension):
        """ Test with double @@ -> replace value with content of content
        """
        generic_temp_folder = Path(generic_temp_folder)

        sample_dict = data['data']
        to_be_raplaced_keys = data['to_be_replaced']  # , _, _, to_be_raplaced_keys = data
        volatile_dict = copy.deepcopy(sample_dict)

        subtitutions_values = {}
        for k in to_be_raplaced_keys:
            random_name = k + f".{cfg_extension}"  # {XConfig.REFERENCE_QUALIFIER}"
            random_name = DirectiveAT.generate_directive_string('import_root', [random_name])
            # random_name = k + f".{cfg_extension}{XConfig.REFERENCE_QUALIFIER}{XConfig.REFERENCE_QUALIFIER}"
            subtitutions_values[random_name] = pydash.get(volatile_dict, k)

            pydash.set_(volatile_dict, k, random_name)

        output_cfg_filename = generic_temp_folder / f'out_config.{cfg_extension}'

        subtitutions_values[str(output_cfg_filename)] = volatile_dict

        saved_cfgs = []
        for directive_value, d in subtitutions_values.items():
            directive = DirectiveAT(value=directive_value)
            if directive.valid:
                output_filename = generic_temp_folder / directive.args[0]
            else:
                output_filename = generic_temp_folder / directive_value
            # output_filename = generic_temp_folder / output_filename.replace(f'{XConfig.REFERENCE_QUALIFIER}', '')
            store_cfg(output_filename, d)
            saved_cfgs.append(output_filename)

        yconf = XConfig(output_cfg_filename)

        assert DeepDiff(yconf.to_dict(), sample_dict)


class TestXConfigReplace(object):

    @pytest.mark.parametrize("data", data_to_test())
    def test_replace(self, generic_temp_folder, data):

        sample_dict = data['data']
        placeholders = data['placeholders']  # , _, _, to_be_raplaced_keys = data
        environment_variables = data['environment_variables']

        conf = XConfig.from_dict(sample_dict)

        np.random.seed(66)

        to_replace = {}
        for p in placeholders:
            _p = Placeholder.from_string(p)
            to_replace[_p.name] = np.random.randint(0, 10)

        assert len(conf.available_placeholders()) == len(placeholders) + len(environment_variables)

        conf.check_available_placeholders(close_app=False)

        if len(conf.available_placeholders()) > 0:

            for name, p in conf.available_placeholders().items():
                assert isinstance(name, str)
                assert isinstance(p, Placeholder)

            with pytest.raises(SystemExit):
                conf.check_available_placeholders(close_app=True)
            conf.replace_variables_map(to_replace)

            chunks = conf.chunks()
            for key, value in chunks:
                if not isinstance(value, Box) and not isinstance(value, BoxList):
                    assert value not in to_replace.keys()

            assert len(conf.available_placeholders()) == len(environment_variables)


class TestXConfigEnvironmentVariable(object):

    @pytest.mark.parametrize("data", data_to_test())
    def test_env_variables(self, generic_temp_folder, data):

        sample_dict = data['data']
        placeholders = data['placeholders']  # , _, _, to_be_raplaced_keys = data
        environment_variables = data['environment_variables']

        conf = XConfig.from_dict(sample_dict)
        np.random.seed(66)

        assert len(conf.available_placeholders()) == len(placeholders) + len(environment_variables)

        for envv in environment_variables:
            pl = Placeholder.from_string(envv)
            assert pl.name not in os.environ, f"this PYTEST needs no environment variable with name '{pl.name}' "

        # Load and parse environment variables but nobody set them!
        conf = XConfig.from_dict(sample_dict, replace_environment_variables=True)
        assert len(conf.available_placeholders()) == len(placeholders) + len(environment_variables)

        # Load and parse environment variables with manual set
        for envv in environment_variables:
            pl = Placeholder.from_string(envv)
            os.environ[pl.name] = str(np.random.uniform(0, 1, (1,)))

        conf = XConfig.from_dict(sample_dict, replace_environment_variables=True)
        assert len(conf.available_placeholders()) == len(placeholders)


class TestXConfigValidate(object):

    @pytest.mark.parametrize("data", data_to_test())
    def test_validation(self, generic_temp_folder, data):

        sample_dict = data['data']
        schema = data['schema']  # , _, _, to_be_raplaced_keys = data

        conf = XConfig.from_dict(sample_dict)
        conf.set_schema(schema)

        print(type(conf.get_schema()))
        conf.validate()
        assert conf.is_valid()

        conf.set_schema(None)
        assert conf.is_valid()

        invalid_schema = Schema({'fake': int})
        conf.set_schema(invalid_schema)
        with pytest.raises(SchemaMissingKeyError):
            conf.validate()
        assert not conf.is_valid()



import rich
import pytest
from schema import Or, SchemaError
from choixe.spooks import Spook


class SimpleUserObject__(Spook):

    def __init__(
        self,
        a: int,
        b: str,
        c: bool
    ) -> None:
        super().__init__()
        self.a = a
        self.b = b
        self.c = c

    def __eq__(self, o: object) -> bool:
        return \
            self.a == o.a and \
            self.b == o.b and \
            self.c == o.c


class CustomUserObject__(Spook):

    def __init__(
        self,
        a: int = None,
        b: str = None,
        c: bool = None
    ) -> None:
        super().__init__()
        self._misc = (a, b, c)

    @classmethod
    def spook_schema(cls) -> dict:
        return {
            'a': Or(None, int),
            'b': Or(None, str),
            'c': Or(None, bool)
        }

    @classmethod
    def from_dict(cls, d: dict):
        return cls(
            a=d['a'],
            b=d['b'],
            c=d['c']
        )

    def to_dict(self) -> dict:
        return {
            'a': self._misc[0],
            'b': self._misc[1],
            'c': self._misc[2],
        }

    def __eq__(self, o: object) -> bool:
        return \
            self._misc[0] == o._misc[0] and \
            self._misc[1] == o._misc[1] and \
            self._misc[2] == o._misc[2]


class NestedUserObject_(Spook):

    def __init__(self, name: str, o0: SimpleUserObject__, o1: SimpleUserObject__):
        self._name = name
        self._o0 = o0
        self._o1 = o1

    @classmethod
    def spook_schema(cls) -> dict:
        return {
            'name': str,
            'o0': dict,
            'o1': dict
        }

    @classmethod
    def from_dict(cls, d: dict):
        return cls(
            name=d['name'],
            o0=Spook.create(d['o0']),
            o1=Spook.create(d['o1'])
        )

    def to_dict(self) -> dict:
        return {
            'name': self._name,
            'o0': self._o0.serialize(),
            'o1': self._o1.serialize(),
        }

    def __eq__(self, o: object) -> bool:
        return \
            self._name == o._name and \
            self._o0 == o._o0 and \
            self._o1 == o._o1


class TestSpookSimpleSerialization(object):

    def test_simple_serialization(self):

        # args_list = [
        #     {'a': 666, 'b': 'hello!', 'c': True},
        #     {'a': None, 'b': None, 'c': None},
        # ]
        items_list = [
            {'kwargs': {'a': 666, 'b': 'hello!', 'c': True}, 'valid': True},
            {'kwargs': {'a': 'error!', 'b': 'hello!', 'c': True}, 'valid': False},
            {'kwargs': {'a': 66.6, 'b': 'hello!', 'c': True}, 'valid': False},
            {'kwargs': {'a': None, 'b': 'hello!', 'c': True}, 'valid': True},
        ]

        for item in items_list:
            kwargs = item['kwargs']
            valid = item['valid']
            o_0 = SimpleUserObject__(**kwargs)
            rep_0 = o_0.serialize()

            o_1 = Spook.create(rep_0)
            assert isinstance(o_1, SimpleUserObject__)
            rep_1 = o_1.serialize()

            assert o_1 == o_0
            print(rep_0, rep_1)


class TestSpookCustomSerialization(object):

    def test_custom_serialization(self):

        items_list = [
            {'kwargs': {'a': 666, 'b': 'hello!', 'c': True}, 'valid': True},
            {'kwargs': {'a': 'error!', 'b': 'hello!', 'c': True}, 'valid': False},
            {'kwargs': {'a': 66.6, 'b': 'hello!', 'c': True}, 'valid': False},
            {'kwargs': {'a': None, 'b': 'hello!', 'c': True}, 'valid': True},
        ]

        for item in items_list:
            kwargs = item['kwargs']
            valid = item['valid']
            o_0 = CustomUserObject__(**kwargs)

            if valid:
                rep_0 = o_0.serialize(validate=True)
            else:
                with pytest.raises(SchemaError):
                    rep_0 = o_0.serialize(validate=True)
                continue

            o_1 = Spook.create(rep_0, validate=True)
            assert isinstance(o_1, CustomUserObject__)
            rep_1 = o_1.serialize(validate=True)
            assert o_1 == o_0


class TestSpookNestedSerialization(object):

    def test_nested_serialization(self):

        items_list = [
            {
                'kwargs': {
                    'name': 'n0',
                    'o0': {'a': 1, 'b': 'hello', 'c': True},
                    'o1': {'a': 2, 'b': 'world!', 'c': False}
                },
                'valid': True
            },
            {
                'kwargs': {
                    'name': 'n0',
                    'o0': {'a': 'error!', 'b': 'world!', 'c': False},
                    'o1': {'a': 2, 'b': 'world!', 'c': False}
                },
                'valid': False
            },
        ]

        for item in items_list:
            kwargs = item['kwargs']
            valid = item['valid']

            o0 = CustomUserObject__(**kwargs['o0'])
            o1 = CustomUserObject__(**kwargs['o1'])
            name = kwargs['name']

            nested_0 = NestedUserObject_(name=name, o0=o0, o1=o1)

            if valid:
                rep_0 = nested_0.serialize(validate=True)
                rich.print("Representaiton", rep_0)
                nested_1 = Spook.create(rep_0)
                assert nested_0 == nested_1
            else:
                with pytest.raises(SchemaError):
                    nested_0.serialize(validate=True)



from enum import Enum, auto
from typing import Union
from choixe.directives import Directive, DirectiveConsumer, DirectiveFactory
from pathlib import Path


class ImporterType(Enum):
    IMPORT = auto()
    IMPORT_ROOT = auto()

    @classmethod
    def values(cls):
        return list([c.name.lower() for c in cls])

    @classmethod
    def get_type(cls, value: str) -> Union['ImporterType', None]:
        return ImporterType[value.upper()]

    # @classmethod
    # def get_type_as_string(cls, value: str) -> str:
    #     tp = cls.get_type(value)
    #     if isinstance(tp, ImporterType):
    #         return tp.name
    #     return 'None'


class Importer(DirectiveConsumer):

    def __init__(self, directive: Directive):
        super().__init__(directive=directive)

    def is_valid(self):
        if self._directive.valid:
            return self._directive.label in ImporterType.values()
        return False

    @property
    def path(self):
        if self.is_valid():
            if len(self._directive.args) > 0:
                return self._directive.args[0]
        return ''

    @property
    def type(self):
        if self.is_valid():
            return ImporterType.get_type(self._directive.label)
        return None

    @classmethod
    def from_string(cls, value: str) -> 'Importer':
        directive = DirectiveFactory.build_directive_from_string(value)
        if directive:
            return Importer(directive=directive)
        return None



from typing import Callable, Sequence
import uuid
import pika
import pickle


class Bulletin(object):

    def __init__(self, metadata: dict = None):
        self.metadata = metadata if metadata is not None else {}

    def to_bytes(self) -> bytes:
        return pickle.dumps(self)

    @classmethod
    def from_bytes(cls, b: bytes) -> 'Bulletin':
        return pickle.loads(b)


class BulletinBoard(object):

    def __init__(self, session_id: str, host: str = 'localhost') -> None:
        super().__init__()

        self._host = host
        self._sessionid = session_id
        self._connection = pika.BlockingConnection(pika.ConnectionParameters(self._host))
        self._channel = self._connection.channel()
        self._channel.queue_declare(queue=self._sessionid)
        self._callbacks: Sequence[Callable] = []
        self._started = False

    def register_callback(self, cb: Callable):
        if len(self._callbacks) == 0:
            self._register_proxy()
        self._callbacks.append(cb)

    def hang(self, bulletin: Bulletin):
        self._channel.basic_publish(
            exchange='',
            routing_key=self._sessionid,
            body=bulletin.to_bytes()
        )

    def _register_proxy(self):
        self._channel.basic_consume(
            queue=self._sessionid,
            auto_ack=True,
            on_message_callback=self._callbacks_proxy
        )

    def _callbacks_proxy(self, ch, method, properties, body):
        bl = Bulletin.from_bytes(body)
        for cb in self._callbacks:
            cb(bl)

    def wait_for_bulletins(self):
        self._started = True
        self._channel.start_consuming()

    def close(self):
        # if self._started:
        self._channel.stop_consuming()
        # self._connection.close()



from dataclasses import field
import os

from numpy.lib.function_base import place
from choixe.importers import Importer, ImporterType
from choixe.placeholders import Placeholder, PlaceholderType
from choixe.directives import DirectiveAT, DirectiveFactory
from enum import Enum, auto
from box.from_file import converters
from box import box_from_file, Box, BoxList
import numpy as np
import pydash
from typing import Any, Dict, List, Sequence, Tuple, Union
from schema import Schema
from pathlib import Path
import re


class XConfig(Box):
    PRIVATE_QUALIFIER = '_'
    REFERENCE_QUALIFIER = '@'
    REPLACE_QUALIFIER = '$'
    KNOWN_EXTENSIONS = converters.keys()
    PRIVATE_KEYS = ['_filename', '_schema']

    def __init__(self, filename: str = None, **kwargs):
        """ Creates a XConfig object from configuration file
        :param filename: configuration file [yaml, json, toml], defaults to None
        :type filename: str, optional
        :param replace_environment_variables: TRUE to auto replace environemnt variables placeholders, defaults to False
        :type replace_environment_variables: bool, optional
        :param plain_dict: if not None will be used as data source instead of filename, defaults to None
        :type plain_dict: dict, optional
        """

        # options
        replace_env_variables = kwargs.get('replace_environment_variables', True)
        _dict = kwargs.get('plain_dict', None)

        self._filename = None

        if _dict is None:
            if filename is not None:
                self._filename = Path(filename)
                self.update(box_from_file(file=Path(filename)))
        else:
            self.update(_dict)

        self._schema = None
        self.deep_parse(replace_environment_variables=replace_env_variables)

    @property
    def root_content(self) -> Union[None, Any]:
        """ Returns the 'value' if configuration is a single 'key'/'value' pair.
        Otherwise None
        :return: 'value' if configuration is a single 'key'/'value' pair. Otherwise None
        :rtype: Union[None, Any]
        """
        d = self.to_dict(discard_private_qualifiers=True)
        if len(d.keys()) == 1:
            return d[list(d.keys())[0]]
        return None

    def get_schema(self):
        return self._schema

    def set_schema(self, s: Schema):
        """ Push validation schema
        :param schema: validation schema
        :type schema: Schema
        """
        if s is not None:
            assert isinstance(s, Schema), "schema is not a valid Schema object!"
        self._schema: Schema = s

    def validate(self):
        """ Validate internal schema if any """
        if self.get_schema() is not None:
            self.get_schema().validate(self.to_dict())

    def is_valid(self) -> bool:
        """ Check for schema validity
        :return: TRUE for valid or no schema inside
        :rtype: bool
        """
        if self.get_schema() is not None:
            return self.get_schema().is_valid(self.to_dict())
        return True

    def save_to(self, filename: str):
        """ Save configuration to output file
        :param filename: output filename
        :type filename: str
        :raises NotImplementedError: Raise error for unrecognized extension
        """
        filename = Path(filename)
        data = self.to_dict()
        if 'yml' in filename.suffix.lower() or 'yaml' in filename.suffix.lower():
            Box(self.decode(data)).to_yaml(filename=filename)
        elif 'json' in filename.suffix.lower():
            Box(self.decode(data)).to_json(filename=filename)
        elif 'toml' in filename.suffix.lower():
            Box(self.decode(data)).to_toml(filename=filename)
        else:
            raise NotImplementedError(f"Extension {filename.suffix.lower()} not supported yet!")

    @classmethod
    def decode(cls, data: any) -> any:
        """ Decode decodable data

        :param data: [description]
        :type data: any
        :return: [description]
        :rtype: any
        """
        if isinstance(data, np.ndarray):
            return cls.decode(data.tolist())
        elif 'numpy' in str(type(data)):
            return cls.decode(data.item())
        elif isinstance(data, list) or isinstance(data, BoxList):
            return [cls.decode(x) for x in data]
        elif isinstance(data, tuple):
            return [cls.decode(x) for x in data]
        elif isinstance(data, dict) or isinstance(data, Box):
            return {k: cls.decode(x) for k, x in data.items()}
        else:
            return data

    def chunks(self, discard_private_qualifiers: bool = True) -> Sequence[Tuple[str, Any]]:
        """ Builds a plain view of dictionary with pydash dot notation
        :param discard_private_qualifiers: TRUE to discard keys starting with private qualifier, defaults to True
        :type discard_private_qualifiers: bool, optional
        :return: list of pairs (key, value) where key is a dot notation pydash key (e.g. d['one']['two']['three'] -> 'one.two.three' )
        :rtype: Sequence[Tuple[str, Any]]
        """
        return self._walk(self, discard_private_qualifiers=discard_private_qualifiers)

    def chunks_as_lists(self, discard_private_qualifiers: bool = True) -> Sequence[Tuple[List[str], Any]]:
        """ Builds a plain view of dictionary with pydash list of str notation
        :param discard_private_qualifiers: TRUE to discard keys starting with private qualifier, defaults to True
        :type discard_private_qualifiers: bool, optional
        :return: list of pairs (key, value) where key is a list of str pydash key (e.g. d['one']['two']['three'] -> ['one', 'two', 'three'] )
        :rtype: Sequence[Tuple[List[str], Any]]
        """
        return self._walk(self, discard_private_qualifiers=discard_private_qualifiers, use_dot_notation=False)

    def is_a_placeholder(self, value: any) -> bool:
        """ Checks if value is likely a placeholder

        :param value: input value to check (e.g. '$hello')
        :type value: any
        :return: TRUE if value is a placeholder
        :rtype: bool
        """
        if not isinstance(value, str):
            return False
        placeholder = Placeholder.from_string(value)
        if placeholder:
            return placeholder.is_valid()
        return False

    def deep_set(self, full_key: str, value: any):
        """ Sets value based on full path key (dot notation like 'a.b.0.d')

        :param full_key: full path key
        :type full_key: str
        :param value: value to set
        :type value: any
        """

        pydash.set_(self, full_key, value)

    def replace_variable(self, old_value: str, new_value: str):
        """ Replaces target variables with custom new value
        :param old_value: value to replace
        :type old_value: str
        :param new_value: new key value
        :type new_value: str
        """
        chunks = self.chunks_as_lists(discard_private_qualifiers=True)
        for k, v in chunks:
            p = Placeholder.from_string(v)
            if p is not None and p.is_valid():
                if old_value == p.name:
                    pydash.set_(self, k, p.cast(new_value))

    def replace_variables_map(self, m: dict):
        """ Replace target old variables with new values represented as dict
        :param m: dict of key/value = old/new
        :type m: dict
        """

        for old_v, new_v in m.items():
            self.replace_variable(old_v, new_v)

    def deep_parse(self, replace_environment_variables: bool = False):
        """ Deep visit of dictionary replacing filename values with a new XConfig object recusively

        :param replace_environment_variables: TRUE to auto replace environment variables
        :type replace_environment_variables: bool
        """
        chunks = self.chunks()
        self._deep_parse_for_importers(chunks)
        if replace_environment_variables:
            self._deep_parse_for_environ(chunks)

    def _deep_parse_for_importers(self, chunks: Sequence[Tuple[str, Any]]):
        """ Deep visit of dictionary replacing filename values with a new XConfig object recusively

        :param chunks: chunks to visit
        :type chunks: Sequence[Tuple[str, Any]]
        :raises NotImplementedError: Importer type not found
        :raises OSError: replace file not found
        """

        for chunk_name, value in chunks:
            if not isinstance(value, str):
                continue
            importer = Importer.from_string(value)
            if importer is not None:
                if importer.is_valid():

                    n_references = value.count(self.REFERENCE_QUALIFIER)

                    p = Path(importer.path)  # Path(value.replace(self.REFERENCE_QUALIFIER, ''))
                    if self._filename is not None and not p.is_absolute():
                        p = self._filename.parent / p

                    if p.exists():
                        self._import_external_file(importer, chunk_name, p)
                    else:
                        raise OSError(f'File {p} not found!')

    def _import_external_file(self, importer: Importer, chunk_name: str, filename: Path):
        """ Imports a generic file into cfg tree

        :param importer: importer directive
        :type importer: Importer
        :param chunk_name: chunk name to replace
        :type chunk_name: str
        :param filename: external filename to import
        :type filename: Path
        :raises NotImplementedError: if importer type is not managed yet
        :raises RuntimeError: if external file content is not readable
        """

        extension = filename.suffix.replace('.', '')
        if extension in self.KNOWN_EXTENSIONS:
            sub_cfg = XConfig(filename=filename)
            if importer.type == ImporterType.IMPORT_ROOT:
                pydash.set_(self, chunk_name, sub_cfg.root_content)
            elif importer.type == ImporterType.IMPORT:
                pydash.set_(self, chunk_name, sub_cfg)
            else:
                raise NotImplementedError(f"Importer type {importer.type} not implemented yet!")
        else:
            try:
                content = open(filename, 'r').read()
                pydash.set_(self, chunk_name, content)
            except UnicodeDecodeError:
                raise RuntimeError(f'Error reading content of file: {str(filename)}. Is this a binary file?')

    def _deep_parse_for_environ(self, chunks: Sequence[Tuple[str, Any]]):
        """ Deep visit of dictionary replacing environment variables if any

        :param chunks: chunks to visit
        :type chunks: Sequence[Tuple[str, Any]]
        :raises NotImplementedError: Importer type not found
        :raises OSError: replace file not found
        """

        env_to_replace = {}
        for chunk_name, value in chunks:
            if not isinstance(value, str):
                continue
            placeholder = Placeholder.from_string(value)
            if placeholder is not None:
                if placeholder.is_valid():
                    if placeholder.type == PlaceholderType.ENV:
                        if placeholder.name in os.environ:
                            env_to_replace[placeholder.name] = os.environ.get(placeholder.name)

        self.replace_variables_map(env_to_replace)

    def to_dict(self, discard_private_qualifiers: bool = True) -> Dict:
        """
        Turn the Box and sub Boxes back into a native python dictionary.
        :return: python dictionary of this Box
        """
        out_dict = dict(self)
        for k, v in out_dict.items():
            # if v is self:
            #     out_dict[k] = out_dict
            if isinstance(v, Box):
                out_dict[k] = self.decode(v.to_dict())
            elif isinstance(v, BoxList):
                out_dict[k] = self.decode(v.to_list())

        if discard_private_qualifiers:
            chunks = self.chunks(discard_private_qualifiers=False)
            for chunk_name, value in chunks:
                c0 = any([x for x in self.PRIVATE_KEYS if x in chunk_name])
                c1 = any([f'.{x}' for x in self.PRIVATE_KEYS if x in chunk_name])
                if c0 or c1:
                    # if chunk_name in self.PRIVATE_KEYS or f'.{chunk_name}' in self.PRIVATE_KEYS:
                    # if f'.{self.PRIVATE_QUALIFIER}' in chunk_name or chunk_name.startswith(self.PRIVATE_QUALIFIER):
                    pydash.unset(out_dict, chunk_name)
        return out_dict

    def available_placeholders(self) -> Dict[str, Placeholder]:
        """ Retrieves the available placeholders list

        :return: list of found (str,str) pairs
        :rtype: Tuple[str,str]
        """

        chunks = self.chunks(discard_private_qualifiers=True)
        placeholders = {}
        for k, v in chunks:
            if self.is_a_placeholder(v):
                placeholders[k] = Placeholder.from_string(v)
        return placeholders

    def check_available_placeholders(self, close_app: bool = False) -> bool:
        """ Check for available placeholder and close app if necessary
        :param close_app: TRUE to close app if at least one placeholder found, defaults to False
        :type close_app: bool, optional
        :return: TRUE if no placeholders found
        :rtype: bool
        """
        placeholders = self.available_placeholders()
        if len(placeholders) > 0:
            import rich
            from rich.table import Table
            from rich.console import Console
            from rich.markdown import Markdown

            console = Console()
            table = Table(show_header=True, header_style='bold magenta')
            table.add_column("Placeholder", style="dim")
            table.add_column("Name")
            table.add_column("Type")
            table.add_column("Options")
            table.add_column("Default")

            header = "*** Incomplete Configuration, Placeholders found! ***"
            rich.print(Markdown(f"# {header}"))

            for k, p in placeholders.items():
                if p.is_valid():
                    table.add_row(k, p.name, p.plain_type, '|'.join(p.options), p.default_value)

            console.print(table)

            if close_app:
                import sys
                sys.exit(1)
            return False
        return True

    @classmethod
    def from_dict(cls, d: dict, **kwargs) -> 'XConfig':
        """ Creates XConfig from a plain dictionary
        : param d: input dictionary
        : type d: dict
        : return: built XConfig
        : rtype: XConfig
        """
        cfg = XConfig(filename=None, plain_dict=d, **kwargs)
        return cfg

    @classmethod
    def _walk(cls,
              d: Dict, path: Sequence = None,
              chunks: Sequence = None,
              discard_private_qualifiers: bool = True,
              use_dot_notation: bool = True) -> Sequence[Tuple[str, Any]]:
        """ Deep visit of dictionary building a plain sequence of pairs(key, value) where key has a pydash notation
        : param d: input dictionary
        : type d: Dict
        : param path: private output value for path(not use), defaults to None
        : type path: Sequence, optional
        : param chunks: private output to be fileld with retrieved pairs(not use), defaults to None
        : type chunks: Sequence, optional
        : param discard_private_qualifiers: TRUE to discard keys starting with private qualifier, defaults to True
        : type discard_private_qualifiers: bool, optional
        : param use_dot_notation: True to use pydash dot notation, False to use list of str notation, defaults to True
        : type use_dot_notation: bool
        : return: sequence of retrieved pairs
        : rtype: Sequence[Tuple[str, Any]]
        """
        root = False
        if path is None:
            path, chunks, root = [], [], True
        if isinstance(d, dict):
            for k, v in d.items():
                path.append(k)
                if isinstance(v, dict) or isinstance(v, list):
                    cls._walk(v, path=path, chunks=chunks, discard_private_qualifiers=discard_private_qualifiers, use_dot_notation=use_dot_notation)
                else:
                    chunk = list(map(str, path))
                    chunk_name = ".".join(chunk)
                    if not(discard_private_qualifiers and chunk_name.startswith(cls.PRIVATE_QUALIFIER)):
                        if use_dot_notation:
                            chunks.append((chunk_name, v))
                        else:
                            chunks.append((chunk, v))
                path.pop()
        elif isinstance(d, list):
            for idx, v in enumerate(d):
                path.append(str(idx))
                cls._walk(v, path=path, chunks=chunks, discard_private_qualifiers=discard_private_qualifiers, use_dot_notation=use_dot_notation)
                path.pop()
        else:
            chunk = list(map(str, path))
            if use_dot_notation:
                chunk = ".".join(chunk)
            chunks.append((chunk, d))
        if root:
            return chunks




from choixe.bulletins import Bulletin, BulletinBoard
import uuid
import threading
import numpy as np
import time
import pytest
import pika

# @pytest.mark.rabbitmq


class TestBulletins(object):

    def test_multi_producer_sinle_consumer(self):

        session_id = str(uuid.uuid1())
        N = 10
        running = True
        try:
            main_board = BulletinBoard(session_id=session_id)
        except pika.exceptions.AMQPConnectionError:
            pytest.skip(msg=f'No rabbitmq server found!')
            return

        for i in range(N):
            board = BulletinBoard(session_id=session_id)
            board.hang(Bulletin(metadata={'value': np.random.randint(0, N)}))

        counter = 0

        def bulletin_update(bulletin: Bulletin):
            nonlocal counter
            assert 'value' in bulletin.metadata
            counter += 1
            print("COUNTER", counter, N)
            if counter >= N:
                main_board.close()

        main_board.register_callback(bulletin_update)
        main_board.wait_for_bulletins()
